package mathx;

/**
 * A matrix class...
 */
class Matrix<T>
{
	private var data : Array<T>;
	public var sizeX(default,null) : Int;
	public var sizeY(default,null) : Int;
	
	public function get( x : Int, y : Int ) : T
	{
		//assert validIndex(x,y);
		return data[ y*sizeX + x ];
	}
	
	/** 
	 * Obtain the value at the given index where if x,y is not valid it will
	 * be constrained to the matrix to get the closest edge value)
	 */
	public function ext_get( x : Int, y : Int ) : T
	{
		if( x < 0 )
			x = 0;
		else if( x >= sizeX )
			x = sizeX - 1;
			
		if( y < 0 )
			y = 0;
		else if( y >= sizeY )
			y = sizeY - 1;
			
		return data[ y*sizeX + x ];
	}
	
	/**
	 * Obtain the value at the index where if x,y is not valid then
	 * a default value will be returned
	 */
	public function def_get( x : Int, y : Int, def : T ) : T	 
	{
		if( x < 0 )
			return def;
		else if( x >= sizeX )
			return def;
			
		if( y < 0 )
			return def;
		else if( y >= sizeY )
			return def;
			
		return data[ y*sizeX + x ];
	}
	
	public function set( x : Int, y : Int, val : T )
	{
		//assert validIndex(x, y );
		data[y*sizeX + x] = val;
	}
	
	static public function validIndexFor( mp : MatPoint, size : MatPoint ) : Bool
	{
		return mp.x >= 0 && mp.x < size.x 
			&& mp.y >= 0 && mp.y < size.y;
	}
	
	static public function validIndexForXY( mpx : Int, mpy : Int, sizex : Int, sizey : Int ) : Bool
	{
		return mpx >= 0 && mpx < sizex 
			&& mpy >= 0 && mpy < sizey;
	}
	
	public function validIndex( x : Int, y : Int ) : Bool
	{
		return x >= 0 && x < sizeX 
			&& y >= 0 && y < sizeY;
	}
	
	//only static functions can create a new matrix
	private function new()
	{
	}
	
	static public function fromArrayMat<Q>( mat : Array<Array<Q>> ) : Matrix<Q>
	{	
		//assert mat.length > 0
		//assert mat[0].length > 0
		
		var m = new Matrix<Q>();
		m.sizeX = mat[0].length;
		m.sizeY = mat.length;
		m.data = new Array<Q>();
		
		for( i in 0...m.sizeY )
		{
			//assert mat[i] == sizeX
			for( j in 0...m.sizeX )
				m.data.push( mat[i][j] );
		}
		
		return m;
	}
	
	/**
	 * Creates a matrix with the default value given
	 */
	static public function create<Q>( cols : Int, rows : Int, init : Q ) : Matrix<Q>
	{
		var m = new Matrix<Q>();
		m.sizeX = cols;
		m.sizeY = rows;
		m.data = new Array<Q>();
		
		for( i in 0...m.sizeY )
			for( j in 0...m.sizeX )
				m.data.push( init );
				
		return m;
	}
	
	/**
	 * Does a copy of the matrix.  The internal items are not copied, as this
	 * is meant for fundamentals (where a copy is always taken
	 */
	public function copy() : Matrix<T>
	{
		var m = new Matrix<T>();
		m.sizeX = sizeX;
		m.sizeY = sizeY;
		m.data = data.copy();
		return m;
	}


	/**
	 * Extracts the row from the matrix
	 */
	public function extractRow( y : Int ) : Array<T>
	{
		//assert bounds
		var r = new Array<T>();
		for( i in 0...sizeX )
			r.push( data[y*sizeX + i] );
		return r;
	}	 
	
	/**
	 * Flips the matrix y[n] <=> y[0].  Returns a new matrix, original is not modified
	 */
	public function createFlip() : Matrix<T>
	{
		var m = new Matrix<T>();
		m.sizeX = sizeX;
		m.sizeY = sizeY;
		m.data = new Array<T>();
		
		for( y in 0...m.sizeY )
			for( x in 0...m.sizeX )
				m.data.push( data[ (m.sizeY-y-1) * sizeX + x ] );
				
		return m;
	}
	
	/**
	 * Produces a standard iterator which iteratres over all matrix
	 * elements starting at 0,0 and incrementing x first (in rows)
	 */
	public function xOrderIter() : MatrixXOrderIter
	{
		return new MatrixXOrderIter( MatPoint.at( sizeX, sizeY ) );
	}
	
	public function subIter( upperLeft : MatPoint, size : MatPoint, ?confine : Null<Bool> ) : MatrixXOrderIter
	{
		//confine to bounds of this matrix
		if( confine == null || confine )
		{
			upperLeft = upperLeft.copy();
			size = size.copy();
			
			if( upperLeft.x < 0 )
			{
				size.x += upperLeft.x;
				upperLeft.x = 0;
			}
			if( upperLeft.y < 0 )
			{
				size.y += upperLeft.y;
				upperLeft.y = 0;
			}
			
			if( upperLeft.x + size.x > sizeX )
				size.x = sizeX - upperLeft.x;	//may result in a negative here...
			if( upperLeft.y + size.y > sizeY )
				size.y = sizeY - upperLeft.y;
							
			//check for empty cases
			if( size.x <= 0 || size.y <= 0 )
			{
				size.x = 0;
				size.y = 0;
			}
			//normal iterator handles such 0 cases then.
		}
		
		return new MatrixXOrderIter( upperLeft.add( size ), upperLeft );
	}
	
	/**
	 * Produces an iterator around the given box.  It starts in the upperleft corner and works
	 * its way clockwise around the region. Only valid matrix points will be retrieved by
	 * this iterator, those lying outside of the matrix will be skipped.
	 *
	 * @param	upperLeft	[in]
	 * @param	size	[in]
	 * @param	radius	[in] how far off of the given segment is the border constructed
	 */
	public function borderIter( upperLeft : MatPoint, size : MatPoint, radius : Int ) : MatrixBorderIter
	{
		var nul = upperLeft.copy();
		nul.x -= radius;
		nul.y -= radius;
		var nsize = size.copy();
		nsize.x += radius * 2;
		nsize.y += radius * 2;
		
		return new MatrixBorderIter( nul, nsize );
	}
	
	/**
	 * Produces a box iterator around the given point.  It will not
	 * return points outside of the matrix, rather just skipping those
	 * when it comes to them
	 */
	public function boxOrderIter( mp : MatPoint ) : MatrixBoxOrderIter
	{
		return new MatrixBoxOrderIter( mp, MatPoint.at( sizeX, sizeY ) );
	}	 
}


///////////////////////////////////////////////////
// Index iterator
//NOTE: a stupid looping is still twice as fast as this!!!
// for( y in 0...matSizeY )
//		for( x in 0...matSizeX )
					
class MatrixXOrderIter /*implements Iterator<MatPoint>*/
{
	var lastAt : MatPoint;	//cache object since object creation can be very slow...
	var nextAtX : Int;
	var nextAtY : Int;
	
	var max : MatPoint;
	var upperLeft : MatPoint;
	
	public function new( amax : MatPoint, ?ul : MatPoint )
	{
		if( ul != null )
			upperLeft = ul.copy();
		else
			upperLeft = MatPoint.at( 0, 0 );
		nextAtX = upperLeft.x;
		nextAtY = upperLeft.y;
		lastAt = upperLeft.copy();
		max = amax.copy();
	}
	
	public function hasNext() : Bool
	{
		return nextAtY < max.y;	//don't need to check x since it is always within range
	}
	
	/** 
	 * The iteration variable itself must be copied if you would like to use it, 
	 * since internally it will reuse the same variable over and over again.
	 */
	public function next() : MatPoint
	{
		lastAt.x = nextAtX;
		lastAt.y = nextAtY;
		
		nextAtX++;
		if( nextAtX == max.x )
		{
			nextAtX = upperLeft.x;
			nextAtY++;
		}
		
		return lastAt;
	}
}

/**
 * The spiral here is a  very blocky one, but good enough for our
 * purposes in Fortress.
 */
class MatrixBoxOrderIter /*implements Iterator<MatPoint>*/
{
	var max : MatPoint;
	var at : MatPoint;
	var radius : Int;	//how large our box is now
	var atD : Int;	//at which diameter are we
	var side : Int;
	
	var nAtX : Int;
	var nAtY : Int;
	var nAtH : Bool;
		
	//were used for dirs array, now just for reference/clarity (side value)
	//static private var bottom = 0;
	//static private var left = 1;
	//static private var top = 2;
	//static private var right = 3;
	
	public function new( src : MatPoint, amax : MatPoint )
	{
		radius = 0;
		at = src.copy();
		max = amax.copy();
		
		nAtX = at.x;
		nAtY = at.y;
		nAtH = true;
	}
	
	public function hasNext() : Bool
	{
		if( !nAtH )
			setNext();
		return nAtH;
	}
	
	public function next() : MatPoint
	{
		if( !nAtH )
			setNext();
			
		at.x = nAtX;
		at.y = nAtY;
		nAtH = false;
		return at;
	}
		
	private function setNext() 
	{
		nAtH = true;
		nAtX = at.x;
		nAtY = at.y;
		
		if( radius == 0 )
		{
			radius = 1;
			atD = 0;
			side = 0;
			nAtX++;	//below will put us at the starting position... (same as side==4 in loop below)
			nAtY--;
		}
		
		do
		{
			//are we done this side
			if( atD == (radius*2) )
			{
				atD = 0;
				side++;
				
				//are we done this layer
				if( side == 4 )
				{
					nAtX++;
					nAtY--;	//move us down-right one
					side = 0;
					radius++;
					
					//are we absolutely done?
					if( radius>=max.x && radius>=max.y)
					{
						//TODO: judging by border loop needing an extra finishing dir, chances
						//are this also needs a special end-case for when hasNext is called again
						//liekly now it produces a very slow check (one full box per hasNext)
						nAtH = false;
						return;
					}
				}
			}
			
			//continue on side (previosly an array was used, but that is far too slow)
			switch( side )
			{
				case 0:
					nAtX--;
				case 1:
					nAtY++;
				case 2:
					nAtX++;
				case 3:
					nAtY--;
			}
			atD++;
			
		//} while( !Matrix.validIndexForXY( nAtX, nAtY, max.x, max.y ) );	//go until our index is valid
		} while( nAtX < 0 || nAtY < 0 || nAtX >= max.x || nAtY >= max.y );	//like above, but avoids function call (which is *very* costly)
		
	}
}


/**
 * The border iterator as described by the BorderIter function
 *
 * TODO: This and the above share the same outer logic, perhaps it could be combined
 * into one pattern -- the onyl fear is that function calls are so costly that this may be
 * a siginificant performance loss (maybe M4 based).
 */
class MatrixBorderIter /*implements Iterator<MatPoint>*/
{
	var nAtX : Int;
	var nAtY : Int;
	var nAtH : Bool;
	var at : MatPoint;
	
	var upperLeft : MatPoint;
	var max : MatPoint;
	
	var dir : Int;	//0=right, 1=down, 2=left, 3=up
	
	public function new( upperLeft : MatPoint, size : MatPoint )
	{
		//TODO: add such support later (it's a special case) (likely add in function, not to this class)
		Assert.isTrue( size.x > 1 && size.y > 1 );
		
		this.upperLeft = upperLeft.copy();
		this.max = upperLeft.add( size );
		
		nAtX = upperLeft.x;
		nAtY = upperLeft.y;
		at = MatPoint.at( nAtX, nAtY );
		nAtH = true;	//first point is the next point
		dir = 0;
	}
	
	public function hasNext() : Bool
	{
		if( !nAtH )
			setNext();
		return nAtH;
	}
	
	public function next() : MatPoint
	{
		if( !nAtH )
			setNext();
			
		at.x = nAtX;
		at.y = nAtY;
		nAtH = false;
		return at;
	}
	
	private function setNext() 
	{
		nAtH = true;
		nAtX = at.x;
		nAtY = at.y;
		
		do 
		{
			switch( dir )
			{
				case 0:	//top-right
					nAtX++;
					if( nAtX == max.x )
						dir++;
						
				case 1:	//right-down
					nAtY++;
					if( nAtY == max.y )
						dir++;
						
				case 2:	//bottom-left
					nAtX--;
					if( nAtX == upperLeft.x )
						dir++;
						
				case 3://left-up
					nAtY--;
					if( nAtY == upperLeft.y )
					{
						dir++;	//set to 4, which will always be false then
						nAtH = false;
						return;
					}
					
				case 4:
					nAtH = false;
					return;					
			}
		} while( nAtX < 0 || nAtY < 0 || nAtX >= max.x || nAtY >= max.y );	//see box order iter
	}
}
