package tutorial;

import flash.events.MouseEvent;

import ui.LinearLayout;
import ui.SpriteWidget;
import ui.Button;
import ui.Action;

import mathx.Matrix;
import mathx.MatPoint;
import mathx.MatrixUtil;

/**
 * Implements our land layout and Wall detection.  IT is a simple
 * grid, on which the user can click to modify the land type 
 * of cells (land, water, wall) Another class may use this widget 
 * and connect buttons to its actions.
 */
class GameField extends SpriteWidget
{
	var fillMethod : Int; //4 to ignore edges, 8 to take edges into account
	
	var cx : Float;	//width of cell
	var cy : Float;	//height of cell
	
	/**
	 * The landMap contains the following cell types
	 */
    var landMap : Matrix<Int>;
    static var ctWater = 0;
	static var ctLand = 1;
	static var ctWall = 2;

	
	public function new()
	{
		super();
		fillMethod = 4;
		
		var sizeX = 40;
		var sizeY = 30;
		
        //create an area of part-land, part-water
		landMap = Matrix.create( sizeX, sizeY, ctLand );
		for( cell in landMap.subIter(MatPoint.at( 0, sizeY - 10 ), MatPoint.at( sizeX, 10)) ) 
		{
			landMap.set( cell.x, cell.y, ctWater );
		}
			
		fillStates = Matrix.create(sizeX,sizeY, fsUnknown );
		
		addEventListener( MouseEvent.MOUSE_DOWN, onMouseDown );	//handle initial down
		addEventListener( MouseEvent.MOUSE_MOVE, onMouse );	//allow drawing while moving the mouse
		addEventListener( MouseEvent.MOUSE_UP, onMouseUp );
	}
	
	/**
	 * This is the drawing function which produces the landmap, and marks the
	 * closed sections with various colours
	 */
	function _resize( w : Float, h : Float )
	{
		cx = w / landMap.sizeX;
		cy = h / landMap.sizeY;
		
		graphics.clear();
		for( cell in landMap.xOrderIter() )
		{
			var t = landMap.get( cell.x, cell.y );
			switch( t )
			{
				case ctWater:
					graphics.beginFill( 0x000080 );
				case ctLand:
					graphics.beginFill( 0x008000 );
				case ctWall:
					graphics.beginFill( 0x808080 );
				default:
					Assert.unreachable();
			}
			graphics.lineStyle( 0, 0x000000 );
			graphics.drawRect( cx * cell.x, cy * cell.y, cx, cy );
			graphics.endFill();
			
			var fillStateOffset = fillStates.get( cell.x, cell.y ) - fillAreaStartingId;
			if( fillStateOffset < 0 )
				continue;
				
			//color each closed area differently
			var colors = [ 0x800000, 0x800080, 0x808000, 0xFF0000, 0xFF00FF, 0xFFFF00 ];
			graphics.lineStyle();
			graphics.beginFill( colors[ fillStateOffset % colors.length ] );
			graphics.drawRect( cx * (cell.x + 0.25), cy * (cell.y +0.25 ),	cx/2, cy/2 );
			graphics.endFill();
		}
	}
	
	/**
	 * Translate a local screen point into an index into the landMap
	 *
	 * @return	[out] the point in the landMap, or null if invalid
	 */
	function getLocalPoint( x : Float, y : Float ) : MatPoint
	{
		var x = Math.floor( x / cx );
		var y = Math.floor( y / cy );
		//coordinates are odd in flash, it is possible to get something outside what we expect, thus this check...
		if( !landMap.validIndex( x, y ) )
			return null;
		return MatPoint.at( x, y );
	}
	
	var assignT : Int;
	
	/**
	 * MouseDown determines what type to be applied, so that the
	 * mouse movements do not continually alter the type
	 */
	function onMouseDown( evt : MouseEvent )
	{
		//coordinates are odd in flash, it is possible to get something outside what we expect, thus this check...
		var at = getLocalPoint( evt.localX, evt.localY );
		if( at == null ) 
			return;
			
		//set to water if Ctrl key pressed, otherwise alternate between land/wall
		if( evt.ctrlKey )
			assignT = ctWater;
		else
		{
			var t = landMap.get( at.x, at.y );
			if( t == ctLand )
				assignT = ctWall;
			else
				assignT = ctLand;
		}
		
		onMouse( evt );
	}
	
	function onMouse( evt : MouseEvent )
	{
		//only apply while the mouse is down
		if( !evt.buttonDown )
			return;
			
		var at = getLocalPoint( evt.localX, evt.localY );
		if( at == null ) 
			return;
			
		if( landMap.get( at.x, at.y ) == assignT )
			return;	//avoid unccessary redrawing
			
		landMap.set( at.x, at.y, assignT  );
				
		evt.updateAfterEvent();
		update();
	}
	
	function onMouseUp( evt : MouseEvent )
	{
		recalc();	//remove if too slow...
	}
	
	/**
	 * Use seedfill4 as the filling algorithm.  This means that diagonals are not considered
	 * connected and thus the "corners" of a wall don't need to be cocelllete.
	 */
	public function setFillMethod4()
	{
		fillMethod = 4;
		recalc();
	}
	
	/**
	 * changes fillMethod to use 8 offsets.  This thus considers all diagonals, so unfinished
	 * corners leave the wall unclosed.
	 */
	public function setFillMethod8()
	{
		fillMethod = 8;
		recalc();
	}
	
    var fillStates : Matrix<Int>;
	static public var fsUnknown = 0;	 
	static public var fsOpen = 1;
	static public var fsStopFill = 2;
	static public var fillAreaStartingId = 3;
	
	/** 
	 * This is the interesting part of the GameField.  Here is where the landMap is
	 * used to calculate the enclosed areas.
	 *
	 * The edges of the map and the shoreline are 'open'. If you wish to treat 
     * the edge of the map or the shoreline as completing a wall, 
     * those 'fsOpen' assignments can be changed to 'fsStopFill'
	 */
	public function recalc()
	{
		var opencells = new FastList<MatPoint>();
		
		//Initialize cell states
		for( cell in landMap.xOrderIter() )
		{
			var cellType = landMap.get(cell.x,cell.y);
			var fillState;
            if( cellType == ctWater || cellType == ctWall)
				fillState = fsStopFill; 
            else if( alongEdge(cell) || cellType == ctLand && onShore( cell ))
				fillState = fsOpen;	
			else
				fillState = fsUnknown;
				
			fillStates.set(cell.x,cell.y, fillState);
			if( fillState == fsOpen )
				opencells.push( cell.copy() );
		}
		
		//Fill the Open areas 
        seedFill( fillStates, opencells, fsUnknown, fsOpen );
		
		//Fill the enclosed areas (fill each with a different identifier)
        var areaId = fillAreaStartingId;
		
		for( cell in landMap.xOrderIter() )
		{
			//anything which is not unknown has already been accounted for
			var fillState = fillStates.get(cell.x,cell.y);
			if( fillState != fsUnknown )
				continue;
				
			//otherwise this cell is part of an enclosed area - fill it with the current areaId
			var enclosedAreaCells = new FastList<MatPoint>();
			enclosedAreaCells.push( MatPoint.at( cell.x, cell.y ) );
            seedFill(fillStates, enclosedAreaCells, fsUnknown, areaId );
				
			areaId++;
		}
		
		//Finally redraw to show what has changed
		update();
	}
    
    /**
     * Perform the right seed fill depending on the fill method that's been set
     */
    function seedFill(colorMap : Matrix<Int>, seeds : FastList<MatPoint>, match : Int,  fill : Int)
    {
        if( fillMethod == 4 )
            MatrixUtil.seedFill4( colorMap, seeds, match, fill );
        else
            MatrixUtil.seedFill8( colorMap, seeds, match, fill );
    }
    
    /**
     * Is the point along the edge of the field?
     */
    function alongEdge(cell : MatPoint) : Bool
    {
        return cell.x == 0 || cell.y == 0 || cell.x == (landMap.sizeX-1) || cell.y == (landMap.sizeY-1);
    }
	
	/**
	 * Determines whether this land is on the shore, that is, it borders
	 * on water.  This simply checks the surrounding cells for water.
	 * Note, the "ext_get" function allows us to lookup cells outside the
	 * size of the grid (in case we are on the borders).  It does so by
	 * pretending the grid extends infinitely in all directions (with the
	 * same value as on the edge)
	 */
	function onShore( cell : MatPoint ) : Bool
	{
		var offsets = (fillMethod == 4) ? MatrixUtil.offsets4 : MatrixUtil.offsets8;
		for( off in offsets )
		{
			var p = cell.add( MatPoint.at( off[0], off[1] ) );
			if( landMap.ext_get( p.x, p.y ) == ctWater )
				return true;
		}		
		return false;
	}
}

/**
 * The main layout class.
 */
class WallDetect extends ui.StageLayout
{
	var layout : LinearLayout;
	var gameField : GameField;
		
	function new()
	{
		super( mathx.MatPoint.at( 40, 20 ) );
		
		layout = new LinearLayout( ui.LayoutDirection.Vertical );
		layout.spacing =0 ;
		addChild( layout );
		
		var buttons = new LinearLayout( ui.LayoutDirection.Horizontal );
		layout.add( buttons );
		 
		gameField = new GameField();
		layout.add( gameField, ui.SizeType.Fill( 1 ) );
		
		buttons.add( new Button( "4 seed", Action.bind( gameField.setFillMethod4 ) ) );
		buttons.add( new Button( "8 seed", Action.bind( gameField.setFillMethod8 ) ) );
	}
	
	function resize( w : Float, h : Float )
	{
		layout.resize( w, h );
	}
	
	static public function main()
	{
		ui.StageLayout.setup( new WallDetect() );
	}
}
