package ui;

typedef RGBSplit = { var r : Int; var g : Int; var b : Int; };
typedef RGBSplitF = { var r : Float; var g : Float; var b : Float; };
typedef HSVF = { var h : Float; var s : Float; var v : Float; };

/**
 * Some functions for colors. Unless otherwise noted the color model being used
 * is RGB.
 */
class ColorUtil
{
	static public function split( color : Int ) : RGBSplit
	{
		return 
			{
			r : (color >> 16) & 0xFF,
			g : (color >> 8) & 0xFF,
			b : color & 0xFF
			}
	}
	
	static public function splitF( color : Int ) : RGBSplitF
	{
		return 
			{
			r : ( (color >> 16) & 0xFF ) / 255.0,
			g : ( (color >> 8) & 0xFF ) / 255.0,
			b : ( color & 0xFF ) / 255.0
			}
	}
	
	static public function combine( rgb : RGBSplit ) : Int
	{
		return (rgb.r << 16) | (rgb.g << 8) | rgb.b;
	}
	
	static public function combineF( rgb : RGBSplitF ) : Int
	{
		return 	
		( Math.round( rgb.r * 255 ) << 16) 
		| ( Math.round( rgb.g * 255) << 8) 
		| Math.round( rgb.b * 255 );
	}
	
	static public function brighten( color : Int ) : Int
	{
		return adjustBrightness( color, 1.20 );
	}
	
	static public function darken( color : Int ) : Int
	{
		return adjustBrightness( color, 1.0 / 1.20 );	//so that brighten restores original color
	}
	
	static public function adjustBrightness( color : Int, factor : Float )
	{
		var rgb = split( color );
		
		rgb.r = Math.floor( Math.min( 255, rgb.r * factor ) );
		rgb.g = Math.floor( Math.min( 255, rgb.g * factor ) );
		rgb.b = Math.floor( Math.min( 255, rgb.b * factor ) );
		
		return combine( rgb );
	}
	
	//algorithm from http://en.wikipedia.org/wiki/HSV_color_space
	static public function rgbToHSV( color : Int ) : HSVF
	{
		var rgb = splitF( color );
		
		var mx = Math.max( Math.max( rgb.r, rgb.g ), rgb.b );
		var mn = Math.min( Math.min( rgb.r, rgb.g ), rgb.b );
		
		var h : Float;
		if( mx == mn )
			h = 0;	//actually undefined.
		else if( mx == rgb.r && rgb.g >= rgb.b )
			h = 60 * (rgb.g-rgb.b) / (mx - mn);
		else if( mx == rgb.r && rgb.g < rgb.b )
			h = 60 * (rgb.g-rgb.b) / (mx - mn) + 360;
		else if( mx == rgb.g )
			h = 60 * (rgb.b-rgb.r) / (mx - mn) + 120;
		else //if( max == rgb.b )
			h = 60 * (rgb.r-rgb.g) / (mx - mn) + 240;
			
		var s : Float;
		if( mx == 0.0 )
			s = 0.0;
		else
			s = 1.0 - (mn/mx);
			
		var v = mx;
		
		return { h : h, s : s, v : v };
	}
	
	static public function hsvToRGB( hsv : HSVF ) : Int
	{
		var Hi = Math.floor( hsv.h / 60 ) % 6;
		var f = hsv.h / 60 - Hi;
		var p = hsv.v * (1 - hsv.s );
		var q = hsv.v * ( 1 - f * hsv.s );
		var t = hsv.v * (1 - ( 1- f )*hsv.s );
		
		var rgb = { r:0.0,g:0.0,b:0.0};
		switch( Hi )
		{
			case 0:
				rgb.r = hsv.v;
				rgb.g = t;
				rgb.b = p;
			case 1:
				rgb.r = q;
				rgb.g = hsv.v;
				rgb.b = p;
			case 2:
				rgb.r = p;
				rgb.g = hsv.v;
				rgb.b = t;
			case 3:
				rgb.r = p;
				rgb.g = q;
				rgb.b = hsv.v;
			case 4:
				rgb.r = t;
				rgb.g = p;
				rgb.b = hsv.v;
			case 5:
				rgb.r = hsv.v;
				rgb.g = p;
				rgb.b = q;
			default:
				Assert.unreachable();
		}
		
		return combineF( rgb );
	}
	
	static public function complement( color : Int ) : Int
	{
		var hsv = rgbToHSV( color );
		var hsvp = hsv;
		
		if( hsv.v >= 180.0 )
			hsvp.v -= 180;
		else
			hsvp.v += 180;
			
		hsvp.s = ( hsv.s * hsv.v ) / ( hsv.v * ( hsv.s - 1 ) + 1 );
					
		hsvp.v  = hsv.v * ( hsv.s - 1 ) + 1;
		
		return hsvToRGB( hsvp );
	}
}
