package ui;

import flash.display.DisplayObject;

// Used by LinearLayout to hold items
class ItemHolder implements Sizeable
{
	public var item : Widget;	//though should be a sprite impelmenting sizeable??
	public var size : SizeType;
	public var visible : Bool;
	public var vertical : Bool;
	
	public function new( item : Widget, size : SizeType, vertical : Bool )
	{
		this.item = item;
		this.size = size;
		this.visible = true;
		this.vertical = vertical;
	}
	
	public function getSizeType() : SizeType
	{
		if( !visible )
			return SizeType.Hidden;
			
		//this needs to be put in GridLayout (see TODO there for merging)
		//this must be done here so that the prefHeight and width is recalculated for a resize event
		if( size == null )
		{
			if( vertical )
				return item.getPrefHeight();
			else
				return item.getPrefWidth();
		}
			
		return size;
	}
	
	public var pOffset : Float;
	public var pSize : Float;
	public function setPlacement( offset : Float, size : Float ) : Void
	{
		pOffset = offset;
		pSize = size;
	}
}

/**
 *  Creates a layout panel in a particular direction.
 *
 * Requires manual Update after additions/modifications.
 */
 class LinearLayout extends flash.display.Sprite, implements Widget
{
	var items : Array<ItemHolder>;
	public var spacing : Float;	//applied as margin around all contained items
	public var padding : Float;	//padding applied to all sides
	public var collapseMargin : Bool;	//do not us spacing on sides, only inter-item spacing
	var background : Widget;
	var vertical : Bool;	//false is Horizontal
	
	public function new( dir : LayoutDirection )
	{
		super();
		this.vertical = dir == LayoutDirection.Vertical;
		spacing = 2;
		padding = 0;
		collapseMargin = false;
		items = new Array<ItemHolder>();
	}
	
	public function setBackground( nb : Widget ) 
	{
		if( background != null )
			removeChild( background.getNative() );
			
		background = nb;
		background.move( 0, 0 );
		addChildAt( background.getNative(), 0 );
	}
	
	public function _resize( w : Float, h : Float )
	{
		//for debugging layout sizing...
		//graphics.clear();
		//graphics.lineStyle( 0, 0 );
		//graphics.drawRect( 0, 0, w, h );
		
		if( background != null )
			background.resize( w, h );
		
		var sizer = new Sizer( spacing, collapseMargin, vertical );
		
		//calc size allotments
		var fullP : Float;
		var fullS : Float;
		//swap for horizontal
		if( vertical )
		{
			fullP = h;
			fullS = if( collapseMargin) w else w - spacing * 2;
		}
		else
		{
			fullP = w;
			fullS = if( collapseMargin ) h else h - spacing * 2;
		}
		
		fullP -= 2 * padding;
		fullS -= 2 * padding;
		
		sizer.size( fullP, fullS, items );
		
		//size items
		for( i in items )
		{
			if( i.pSize == 0 )
			{
				i.item.getNative().visible = false;
				continue;
			}
				
			if( vertical )
			{
				i.item.move( padding + if( collapseMargin ) 0 else spacing, padding + i.pOffset );
				i.item.resize( fullS, i.pSize );
			}
			else
			{
				i.item.move( padding + i.pOffset, padding + if( collapseMargin ) 0 else spacing );
				i.item.resize( i.pSize, fullS );
			}
			
			i.item.getNative().visible = true;
		}
	}
	
	/**
	 * Adds an item to the end of the layout
	 *
	 * @param	item	[in] the widget to add
	 * @param	size	[in] how to size the widget, if not specified the widget must have
	 *		a preferred sizing in the given direction
	 */
	public function add( item : Widget, ?size : SizeType )
	{
		insert( item, size, items.length );
	}
	
	/**
	 * Inserts an item into the layout.
	 *
	 * @param	item	[in] the widget to add
	 * @param	size	[in] how to size the widget
	 * @param	pos	[in] at which index to insert (null indicates first )
	 */
	public function insert( item : Widget, size : SizeType, ?pos : Null<Int> )
	{
		if( pos == null )
			pos = 0;
			
		Assert.notNull( item );
		Assert.isTrue( pos >= 0 && pos <= items.length );
		
		items.insert( pos, new ItemHolder( item, size, vertical ) );
		item.getNative().visible = false;
		addChild( item.getNative() );
	}
	
	public function setItemVisible( item : Widget, vis : Bool )
	{
		for( i in items )
			if( i.item == item )
				i.visible = vis;
	}
	
	public function getPrefWidth() : ui.SizeType
	{
		if( vertical )
		{
			var st = SizeType.Hidden;
			for( i in items )
				st = SizeType.Max( st, i.item.getPrefWidth() );
			if( !collapseMargin )
				st = SizeType.Sum( st, SizeType.Fixed( 2 * spacing ) );
			return st;
		}
		else
		{
			var sizer = new Sizer( spacing, collapseMargin, vertical );
			return sizer.recommend( items );
		}
	}
	
	public function getPrefHeight() : ui.SizeType
	{
		if( vertical )
		{
			var sizer = new Sizer( spacing, collapseMargin, vertical );
			return sizer.recommend( items );
		}
		else
		{
			var st = SizeType.Hidden;
			for( i in items )
				st = SizeType.Max( st, i.item.getPrefHeight() );
			if( !collapseMargin )
				st = SizeType.Sum( st, SizeType.Fixed( 2 * spacing ) );
			return st;
		}
	}
	
	
	/**
 * Provides the functionality to create a Widget of a Flash Sprite.
 */	
 
	var sizeW : Float;	//our desired size (may not be the same as width)
	var sizeH : Float;
	var enabled : Bool;
		
	/**
	 * Serves as a Ctor and must be called by the including class during its
	 * own construction.
	 */
	function MixinWidget()
	{
		sizeW = 0;
		sizeH = 0;
		enabled = true;
	}
	
	public function move( x : Float, y : Float ) : Void
	{
		this.x = x;
		this.y = y;
	}
	
	//Note that users of this mixin override the _resize function, not "resize" directly...
	public function resize( w : Float, h : Float ) : Void
	{
		sizeW = w;
		sizeH = h;
		_resize( w, h );
	}
	
	public function update( ) : Void
	{
		_resize( sizeW, sizeH );
	}
	
	public function enable( set : Bool ) : Void
	{
		enabled = set;
		update();
	}

	public function getNative() : flash.display.DisplayObjectContainer
	{
		return this;
	}
	
	
}
