package ui;

import flash.display.DisplayObject;
import flash.display.DisplayObjectContainer;
import flash.events.Event;
import flash.events.MouseEvent;
import flash.events.KeyboardEvent;

/**
 * Executes an display as a modal display.  This will be modal to the
 * program as a whole (the stage) such that no other window, other
 * than what is provided here, can be accessed.
 *
 * Note that exactly when the completionAction is sent cannot be guaranteed as there
 * are many minor paths by which the modal can end.  In particular you cannot assume
 * that when it is done that something like a Popup menu has processed, or even yet
 * dispatched the associated menu event.  Though other classes may choose to provide
 * guarantees (like Dialog)
 */
class Modal
{
	// Parent will be covered with a gray box during processing (only makes sense when Stage is parent)
	var grayParent : Bool;
	var grayParentObj : SimpleBox;
	
	static var modalEvents = [ 
		MouseEvent.MOUSE_DOWN, MouseEvent.MOUSE_UP, 
		MouseEvent.CLICK, MouseEvent.DOUBLE_CLICK,
		MouseEvent.MOUSE_WHEEL,
		
		//In a decent system these should be fine since they shouldn't change the display tree
		//MouseEvent.MOUSE_MOVE, MouseEvent.MOUSE_OVER, MouseEvent.MOUSE_OUT,
		//MouseEvent.ROLL_OUT, MouseEvent.ROLL_OVER,
		
		KeyboardEvent.KEY_DOWN, KeyboardEvent.KEY_UP
		];
		
	/** 
	 * Constructs a default Modal handler.
	 *
	 * grayParent = true
	 */
	public function new()
	{
		grayParent = true;
		grayParentObj = new SimpleBox( Brush.solid(0x808080) );
		//grayParentObj.alpha = 0.5;
		grayParentObj.blendMode = flash.display.BlendMode.MULTIPLY;
	}
	
	var dObject : Window;
	var dParent : DisplayObjectContainer;
	var active : Bool;
		
	/**
	 * Executres the obj in a Modal faction.
	 *
	 * @param	parent [in] of which object this will be a child, 0 indicates to use the stage
	 * @param	obj [in] which window to display
	 */
	public function exec( parent : DisplayObjectContainer, obj : Window )
	{
		Assert.isTrue( !active );
		
		dObject = obj;
		dParent = if( parent == null ) flash.Lib.current.stage else parent;
		
		//setup listeners to know when we are (planned or otherwise)
		dObject.getNative().addEventListener( Event.REMOVED, onRemoved );
		dObject.setCloseListener( Action.bind( onClose ) );
		
		//gray of parent if desired
		if( grayParent )
		{
			grayParentObj.resize( dParent.width, dParent.height );
			dParent.addChild( grayParentObj );
		}
		
		//finally add target window to parent
		dParent.addChild( dObject.getNative() );
		dObject.exec( dParent );
		active = true;
		
		//then setup the events to make use Modal in nature
		//just guess that 100 is a high enough priority to override other events...?
		for( evt in modalEvents )
			flash.Lib.current.stage.addEventListener( evt, capturePhase, true, 100 );
		
		/*as nothing has the focus often the keyboard events start at the stage, and thus
		 *do not go through a capture stage (as they are at their target).  For this reason
		 * we also need to listen in non-capture mode as well.
		 * NOTE: as we assume the stage is never empty this is not true for mouse events
		 */
		flash.Lib.current.stage.addEventListener( KeyboardEvent.KEY_DOWN, capturePhase, false, 100 );
		flash.Lib.current.stage.addEventListener( KeyboardEvent.KEY_UP, capturePhase, false, 100 );
	}
	
	function capturePhase( evt : Event )
	{
		//do before further propogation (so that focus in modal doesn't block the escape)
		if( evt.type == KeyboardEvent.KEY_UP )
		{
			if( cast( evt, KeyboardEvent ).keyCode == flash.ui.Keyboard.ESCAPE ) 
			{
				evt.stopPropagation();
				goEnd(false);	//cancel the menu
				return;
			}
		}
		
		//children of the display object can process such items
		if( dObject.getNative().contains( cast( evt.target, DisplayObject ) ) )
			return;
		
		//stop it
		evt.stopPropagation();
	}
	
	function onRemoved( evt : Event )
	{
		if( evt.target != dObject )
			return;
			
		//in case something else removes us
		goEnd(true);
	}
	
	function onClose() 
	{
		goEnd( false );
	}
	
	function goEnd( removed : Bool )
	{
		if( !active )
			return;
			
		if( grayParent )
			dParent.removeChild( grayParentObj );
			
		//set active first to false to prevent removal recursion
		active = false;
		
		for( evt in modalEvents )
			flash.Lib.current.stage.removeEventListener( evt, capturePhase, true );
		flash.Lib.current.stage.removeEventListener( KeyboardEvent.KEY_DOWN, capturePhase, false );
		flash.Lib.current.stage.removeEventListener( KeyboardEvent.KEY_UP, capturePhase, false );
		dObject.getNative().removeEventListener( Event.REMOVED, onRemoved );
	
		//root popup needs to remove itself (ensure parent hasn't changed however)
		if( !removed && dObject.getNative().parent == dParent )
			dParent.removeChild( dObject.getNative() );
			
		dObject = null;
		var p = dParent;
		dParent = null;
		
		//signal completion (NOTE: Do absolutely last so this modal could technically be used again)
		if( completionAction != null )
			Action.handle( p, completionAction );
	}
	
	
	/**
	 * Allows an action to be executed when the Modal is completed.
	 */
	var completionAction : Dynamic;
	public function setCompletionAction( action : Dynamic )
	{
		completionAction = action;
	}
}