package ui;

import flash.display.DisplayObject;
import flash.display.Sprite;
import flash.events.Event;
import flash.events.MouseEvent;
import flash.events.KeyboardEvent;
import flash.display.DisplayObjectContainer;
import flash.filters.BitmapFilter;
import flash.filters.DropShadowFilter;
import flash.geom.Point;

/**
 * A popup meu.  Items may be added with associated actions.
 *
 * NOTE: For actions based on messaging the event will be posted
 * to the *parent* of the Popup, not the popup itself (since it may no
 * longer be in the display chain at that time)
 */
class Popup extends Sprite, implements Window
{
	private var items : Array<PopupItem>;
	private var activeSub : Popup;
	private var active : Bool;
	private var rootP : Bool;
	
	public function new( )
	{
		super();
		
		items = new Array<PopupItem>();
			
		addEventListener( Event.REMOVED, onRemoved,false, 0, true /*weak*/ );
		
		//give ourselves a drap-shadow
		var a = new Array<BitmapFilter>();
		a.push( new DropShadowFilter() );
		filters = a;
		
		active = false;
	}
	
	private var popParent : Popup;

	private function goSub( aparent : Popup )
	{
		_go( aparent, false );
		popParent = aparent;
	}
	
	/**
	 * Returns whether or not the popup is currently functioning.
	 */
	public function isActive() : Bool
	{
		return active;
	}
	
	public function getNumItems() : Int
	{
		return items.length;
	}
	
	function onRemoved( evt : Event )
	{
		if( evt.target != this || popParent != null )
			return;
			
		//in case something else removes us
		goEnd( true );
	}
	
	function goEnd( removed : Bool )
	{
		if( !isActive() )
			return;
			
		//set active first to false to prevent removal recursion
		active = false;
		var p = parent;
		if( !removed )
		{
			Assert.notNull( parent );
			parent.removeChild( this );
		}
		
		//call the cleanup function (if needed)
		if( closeHandler != null )
			Action.handle( null, closeHandler );	//use null for parent to  make explicit the danger!
		closeHandler = null;
	}
	
	public function go( aparent : DisplayObjectContainer )
	{
		_go( aparent, true );
	}
	
	/**
	 * Executes the menu within the given parent and at the
	 * desired location.
	 *
	 * @param parent [*] to which parent this will be attached
	 * @param ax,ay [in] where in the parent to place the menu
	 */
	public function _go( aparent : DisplayObjectContainer, asRoot : Bool )
	{
		rootP = asRoot;
		active = true;
		
		cleanSub();	 //to handle the case where it pops up somewhere else while in a sub-menu
		popParent = null;
		graphics.clear();
		
		//get max size first
		var maxW = 0.0;
		for( item in items )
			maxW = Math.max( maxW, item.width );
		
		var atY = 4.0;
		for( item in items )
		{
			item.widen( maxW );
			item.y = atY;
			item.x = 3;
			atY += item.height;
		}
		
		graphics.beginFill( 0x008080 );
		graphics.lineStyle( 1, 0x000000 );
		graphics.drawRoundRect( 0, 0, maxW+6, atY + 3, 6 );
		graphics.endFill();
		
		//initial placement
		if( parent != aparent )	//handle case used as Modal popup where Modal sets parent
			aparent.addChild( this );
		
		//adjust placement to ensure it is within the stage
		var ul = aparent.localToGlobal( new Point( x, y ) );
		var br = localToGlobal( new Point( width, height ) );
		if( br.y >= stage.stageHeight )
			ul.y = stage.stageHeight - (br.y - ul.y);
		if( br.x >= stage.stageWidth )
			ul.x = stage.stageWidth - (br.x - ul.x );
					
		var nul = aparent.globalToLocal( ul );
		x = nul.x;
		y = nul.y;
	}
	
	/**
	 * Adds a textual item to the menu.  The event parameter here can either
	 * be a proper event (which will be cloned on dispatch) or a string, which
	 * will then be wrapped in MenuEvent object.
	 *
	 * Refer to MenuEvent on how to catch multiple menu items with one handler.
	 *
	 * @param	text	[in] text to display
	 * @param	event [in] event to dispatch when clicked, either:
	 *		- a string
	 *		- a complete Event object
	 *		- a function to call (via addItemF with A FunctionWrapper)
	 */
	public function addItem( text : String, event : Dynamic )
	{
		event = Action.translate( event );
		var item = new PopupItemString( this, text, event );
		items.push( item );
		addChild( item );
	}
	
	/**
	 * Adds a separator to the menu
	 */
	public function addSeparator( )
	{
		var item = new PopupItemSeparator( this );
		items.push( item );
		addChild( item );
	}
	
	/**
	 * Adds a sub-menu to this menu 
	 *
	 * @param text [in] text to display in this menu
	 * @param sub [*] popup menu to attach to this item
	 */ 
	public function addSubMenu( text : String, sub : Popup )
	{
		var item = new PopupItemSub( this, text, sub );
		items.push( item );
		addChild( item );
	}
	
	/**
	 * Meant to be called by the items in the menu,
	 * this activates the given event -- indicates
	 * that it was clicked
	 *
	 * @return [out] true if menu is dismissed, false otherwise
	 */
	public function clickItem( event : Dynamic ) : Bool
	{
		//delegate to parent, root popup will handle events
		if( popParent != null )
			return popParent.clickItem( event );
		
		if( event == null )
			return false;
		
		cleanSub();
		
		//close this first, and perform action in parent
		var p = parent;
		goEnd( false );
		Action.handle( p, event );
		
		return true;
	}
	
	/**
	 * Hides the popup sub menu if it is currently being
	 * displayed.  This is recursive to those menus to
	 * ensure their children are also removed -- otherwise
	 * when we first popup this sub again its popup sub
	 * will also still be displayed.
	 */
	private function cleanSub( )
	{
		if( activeSub != null )
		{
			activeSub.cleanSub( );
			removeChild( activeSub );
			activeSub = null;
		}
	}
	
	/**
	 * Indicates when an item has been entered in the
	 * menu.  This is called by the individual items in
	 * the menu.  It is mainly used for popup control
	 */
	public function enterItem( item : PopupItem )
	{
		cleanSub();
		
		activeSub = item.getSubMenu();
		if( activeSub == null )
			return;
			
		activeSub.x = item.x + item.width;
		activeSub.y = item.y;
		activeSub.goSub( this );
	}
	
	public function leaveItem( item : PopupItem )
	{
		//don't do anything, leave submenu up since going into submenu will cause this
		//to disappear
	}
	
	////////////////////////////////////////
	// Window
	private var closeHandler : Dynamic;
	
	public function exec( parent : DisplayObjectContainer ) : Void
	{
		go( parent );
	}
	
	public function setCloseListener( func : Dynamic ) : Void
	{
		closeHandler = func;
	}

	public function getNative() : DisplayObjectContainer
	{
		return this;
	}
}
