package ui;

import flash.display.Sprite;
import flash.events.Event;
import flash.events.ProgressEvent;
import flash.display.Loader;
import flash.net.URLRequest;

/**
 * A wrapper class to use a given URL as a display object.
 */
class URLSprite extends Sprite, implements Widget
{
	var loader : Loader;
	var loaded : flash.display.DisplayObject;
	var tback : Widget; //a temporary background while loading, on start, or on error
	var useTBack : Bool;
	
	/**
	 * @param	url [in] the url of the image to display, null if nothing at creation time
	 * @param	tback	[in] true (default) to show an image on loading/error, false to display nothing
	 */
	public function new( url : String, ?tback : Null<Bool> )
	{
		super();
		MixinWidget();
		
		useTBack = (tback == null) ? true : tback;
		
		loadURL( url );
	}
	
	/**
	 * Sets a new URL for the display in this object.
	 *
	 * TODO: if this is called before the previous one is finished the events from
	 * the previous one will still be triggered and likely cause an error.
	 */
	public function loadURL( url : String )
	{
		//remove existing children
		if( loaded != null )
		{
			removeChild( loaded );
			loaded = null;
		}
		disableTBack();
			
		if( url != null )
		{
			loader = new Loader();
			loader.contentLoaderInfo.addEventListener( Event.COMPLETE, onComplete );
			loader.contentLoaderInfo.addEventListener( flash.events.IOErrorEvent.IO_ERROR, onError );
			loader.contentLoaderInfo.addEventListener( flash.events.SecurityErrorEvent.SECURITY_ERROR, onError );
			loader.contentLoaderInfo.addEventListener( ProgressEvent.PROGRESS, onProgress );
			loader.load( new URLRequest( url ) );
		}
		
		if( useTBack )
			enableTBack();
			
		//force update now
		update();
	}
	
	function enableTBack()
	{
		if( tback == null )
		{
			//old drawing of loading...
			//tback = new ui.DrawBox( drawLoading );
			tback = new CircularProgress( 0x00a000 );
			addChild( tback.getNative() );
		}
	}
	
	function disableTBack()
	{
		if( tback != null )
		{
			removeChild( tback.getNative() );
			tback = null;
		}
	}
	
	function onComplete( evt : Event )
	{
		disableTBack();		
		
		//one would suspect using the loader.content directly might be more managable,
		//but a stupid security error gets in the way, somehow loader.content is considered
		//scripting access and flags the sandbox error, so we live with just making the loader
		//itself a child (adding a server proxy or domain files to all other servers would make
		//direct use possible, but since this seems to work normally there is no need)
		loaded = loader;	//.content;
		addChild( loaded );
		
		update();
	}
	
	function onError( evt : Event )
	{
		//always show an error display, even if loading graphic not requested...
		disableTBack();
		tback = new ui.DrawBox( drawError );
		addChild( tback.getNative() );
		update();
	}
	
	function onProgress( evt : ProgressEvent )
	{
		if( !Std.is( tback, CircularProgress ) )
			return;
		var cp = cast( tback, CircularProgress );
		cp.setTotal( evt.bytesTotal );
		cp.setCurrent( evt.bytesLoaded );
	}
	
	function _resize( w : Float, h : Float )
	{
		if( loaded != null )
		{
			loaded.width = w;
			loaded.height = h;
		}
		
		if( tback != null )
			tback.resize( w, h );
	}
	
	static function drawLoading( graphics : flash.display.Graphics, w : Float, h : Float )
	{
		var lw = w / 10;
		graphics.lineStyle( lw, 0x008000, 0.5, true );
		var off = 0.1;
		var eoff = 1 - 2 * off;
		graphics.drawEllipse( w * off, h * off, w * eoff, h * eoff );
	}
	
	static function drawError( graphics : flash.display.Graphics, w : Float, h : Float )
	{
		var lw = w / 10;
		graphics.lineStyle( lw, 0x800000, 0.5, true );
		var off = 0.1;
		var eoff = 1 - 2 * off;
		graphics.moveTo( w * off, h * off );
		graphics.lineTo( w * eoff, h * eoff );
		graphics.moveTo( w * eoff, h * off );
		graphics.lineTo( w * off, h * eoff );
	}
	
	/**
 * Provides the functionality to create a Widget of a Flash Sprite.
 */	
 
	var sizeW : Float;	//our desired size (may not be the same as width)
	var sizeH : Float;
	var enabled : Bool;
		
	/**
	 * Serves as a Ctor and must be called by the including class during its
	 * own construction.
	 */
	function MixinWidget()
	{
		sizeW = 0;
		sizeH = 0;
		enabled = true;
	}
	
	public function move( x : Float, y : Float ) : Void
	{
		this.x = x;
		this.y = y;
	}
	
	//Note that users of this mixin override the _resize function, not "resize" directly...
	public function resize( w : Float, h : Float ) : Void
	{
		sizeW = w;
		sizeH = h;
		_resize( w, h );
	}
	
	public function update( ) : Void
	{
		_resize( sizeW, sizeH );
	}
	
	public function enable( set : Bool ) : Void
	{
		enabled = set;
		update();
	}

	public function getNative() : flash.display.DisplayObjectContainer
	{
		return this;
	}
	

	public function getPrefWidth() : ui.SizeType
	{
		return null;
	}
	
	public function getPrefHeight() : ui.SizeType
	{
		return null;
	}
	
}
